# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import os
import random
import string
import subprocess
from contextlib import closing
from pathlib import Path
from shutil import copytree
from typing import Any

import docker
import pytest

here = Path(__file__).parent

UNIT_TESSERACT_PATH = here / ".." / "examples"
UNIT_TESSERACTS = [Path(tr).stem for tr in UNIT_TESSERACT_PATH.glob("*/")]


def pytest_addoption(parser):
    parser.addoption(
        "--always-run-endtoend",
        action="store_true",
        dest="run_endtoend",
        help="Never skip end-to-end tests",
        default=None,
    )
    parser.addoption(
        "--skip-endtoend",
        action="store_false",
        dest="run_endtoend",
        help="Skip end-to-end tests",
    )


def pytest_collection_modifyitems(config, items):
    """Ensure that endtoend tests are run last (expensive!)."""
    # Map items to containing directory
    dir_mapping = {item: Path(item.module.__file__).parent.stem for item in items}

    # Sort items based on directory
    sorted_items = sorted(items, key=lambda item: dir_mapping[item] == "endtoend_tests")
    items[:] = sorted_items

    # Add skip marker to endtoend tests if not explicitly enabled
    # or if Docker is not available
    def has_docker():
        try:
            docker.from_env().close()
            return True
        except Exception:
            return False

    run_endtoend = config.getvalue("run_endtoend")

    if run_endtoend is None:
        # tests may be skipped if Docker is not available
        run_endtoend = has_docker()
        skip_reason = "Docker is required for this test"
    elif not run_endtoend:
        skip_reason = "Skipping end-to-end tests"

    if not run_endtoend:
        for item in items:
            if dir_mapping[item] == "endtoend_tests":
                item.add_marker(pytest.mark.skip(reason=skip_reason))


@pytest.fixture(scope="session")
def unit_tesseract_names():
    """Return all unit tesseract names."""
    return UNIT_TESSERACTS


@pytest.fixture(scope="session", params=UNIT_TESSERACTS)
def unit_tesseract_path(request) -> Path:
    """Parametrized fixture to return all unit tesseracts."""
    # pass only tesseract names as params to get prettier test names
    return UNIT_TESSERACT_PATH / request.param


@pytest.fixture(scope="session")
def dummy_tesseract_location():
    """Return the dummy tesseract location."""
    return here / "dummy_tesseract"


@pytest.fixture
def dummy_tesseract_package(tmpdir, dummy_tesseract_location):
    """Create a dummy tesseract package on disk for testing."""
    copytree(dummy_tesseract_location, tmpdir, dirs_exist_ok=True)
    return Path(tmpdir)


@pytest.fixture
def dummy_tesseract_module(dummy_tesseract_package):
    """Create a dummy tesseract module for testing."""
    from tesseract_core.runtime.core import load_module_from_path

    return load_module_from_path(dummy_tesseract_package / "tesseract_api.py")


@pytest.fixture
def dummy_tesseract(dummy_tesseract_package):
    """Set tesseract_api_path env var for testing purposes."""
    from tesseract_core.runtime.config import get_config, update_config

    orig_config_kwargs = {}
    orig_path = get_config().tesseract_api_path
    # default may have been used and tesseract_api.py is not guaranteed to exist
    # therefore, we only pass the original path in cleanup if not equal to default
    if orig_path != Path("tesseract_api.py"):
        orig_config_kwargs |= {"tesseract_api_path": orig_path}
    api_path = Path(dummy_tesseract_package / "tesseract_api.py").resolve()

    try:
        # Configure via envvar so we also propagate it to subprocesses
        os.environ["TESSERACT_API_PATH"] = str(api_path)
        update_config()
        yield
    finally:
        # As this is used by an auto-use fixture, cleanup may happen
        # after dummy_tesseract_noenv has already unset
        if "TESSERACT_API_PATH" in os.environ:
            del os.environ["TESSERACT_API_PATH"]
        update_config(**orig_config_kwargs)


@pytest.fixture
def dummy_tesseract_noenv(dummy_tesseract_package):
    """Use without tesseract_api_path to test handling of this."""
    from tesseract_core.runtime.config import get_config, update_config

    orig_api_path = get_config().tesseract_api_path
    orig_cwd = os.getcwd()

    # Ensure TESSERACT_API_PATH is not set with python os
    if "TESSERACT_API_PATH" in os.environ:
        del os.environ["TESSERACT_API_PATH"]

    try:
        os.chdir(dummy_tesseract_package)
        update_config()
        yield
    finally:
        update_config(tesseract_api_path=orig_api_path)
        os.chdir(orig_cwd)


@pytest.fixture
def free_port():
    """Find a free port to use for HTTP."""
    import socket
    from contextlib import closing

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("localhost", 0))
        return s.getsockname()[1]


@pytest.fixture(scope="session")
def docker_client():
    with closing(docker.from_env()) as client:
        yield client


@pytest.fixture
def dummy_image_name(docker_client):
    """Create a dummy image name, and clean up after the test."""
    image_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=16))
    image_name = f"tmp_tesseract_image_{image_id}"
    try:
        yield image_name
    finally:
        if os.environ.get("TESSERACT_KEEP_BUILD_CACHE", "0").lower() not in (
            "1",
            "true",
        ):
            try:
                docker_client.images.remove(image_name, noprune=False, force=True)
            except (docker.errors.ImageNotFound, docker.errors.NotFound):
                pass


@pytest.fixture(scope="module")
def shared_dummy_image_name(docker_client):
    """Create a dummy image name, and clean up after all tests."""
    image_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=16))
    image_name = f"tmp_tesseract_image_{image_id}"
    try:
        yield image_name
    finally:
        if os.environ.get("TESSERACT_KEEP_BUILD_CACHE", "0").lower() not in (
            "1",
            "true",
        ):
            try:
                docker_client.images.remove(image_name, noprune=False, force=True)
            except (docker.errors.ImageNotFound, docker.errors.NotFound):
                pass


@pytest.fixture
def mocked_docker(monkeypatch):
    """Mock docker client."""
    from tesseract_core.sdk import engine

    class MockedContainer(docker.models.containers.Container):
        """Mock Container class."""

        def __init__(self, return_args: dict):
            self.return_args = return_args

        def wait(self, **kwargs: Any):
            """Mock wait method for Container."""
            return {"StatusCode": 0, "Error": None}

        @property
        def attrs(self):
            """Mock attrs method for Container."""
            return {"Config": {"Env": ["TESSERACT_NAME=vectoradd"]}}

        def logs(self, stderr=False, stdout=False, **kwargs: Any):
            """Mock logs method for Container."""
            out = []
            if stdout:
                out.append(json.dumps(self.return_args).encode("utf-8"))
            if stderr:
                out.append(b"hello tesseract")
            return b"\n".join(out)

        def remove(self, **kwargs: Any):
            """Mock remove method for Container."""
            pass

    class MockedDocker:
        """Mock DockerClient class."""

        def close(self):
            """Mock close method for DockerClient."""
            pass

        def info(self):
            """Mock info method for DockerClient."""
            pass

        class images:
            """Mock images subclass."""

            @staticmethod
            def get(image_id: str):
                """Mock get method for images."""
                return MockedDocker.images.list()[0]

            @staticmethod
            def list() -> list[docker.models.images.Image]:
                return [
                    docker.models.images.Image(
                        attrs={
                            "Id": "sha256:123456789abcdef",
                            "RepoTags": ["vectoradd:latest"],
                            "Size": 123456789,
                            "Config": {"Env": ["TESSERACT_NAME=vectoradd"]},
                        },
                    ),
                    docker.models.images.Image(
                        attrs={
                            "Id": "sha256:48932484029303",
                            "RepoTags": ["hello-world:latest"],
                            "Size": 43829489032,
                            "Config": {"Env": ["PATH=/fake-path"]},
                        },
                    ),
                ]

        class containers:
            """Mock containers subclass."""

            @staticmethod
            def run(**kwargs: Any) -> bytes:
                """Mock run method for containers."""
                container = MockedContainer(kwargs)
                if kwargs.get("detach", False):
                    return container
                return container.logs()

            @staticmethod
            def list(**kwargs: Any) -> list[MockedContainer]:
                return [MockedContainer({"TESSERACT_NAME": "vectoradd"})]

    class MockedAPIClient:
        """Mock APIClient class."""

        def close(self):
            """Mock close method for APIClient."""
            pass

        def prune_builds(self, all: bool, filters: dict):
            """Mock prune_builds method for APIClient."""
            pass

    created_ids = set()

    def mocked_subprocess_run(*args, **kwargs):
        """Mock subprocess.run."""
        if "compose" in args[0] and "up" in args[0]:
            # Extract the tesseract id from the command and store it
            for arg in args[0]:
                if "tesseract-" in arg:
                    created_ids.add(arg)

        if "compose" in args[0] and "ls" in args[0]:
            # Return the list of created tesseract ids
            return subprocess.CompletedProcess(
                args=args,
                returncode=0,
                stderr=b"",
                stdout=str.encode(json.dumps([{"Name": pid} for pid in created_ids])),
            )

        if "compose" in args[0] and "down" in args[0]:
            # Remove the tesseract id from the list of created tesseracts
            for arg in args[0]:
                if "tesseract-" in arg:
                    created_ids.remove(arg)

        return subprocess.CompletedProcess(
            args=args, returncode=0, stderr=b"", stdout=b""
        )

    def mock_from_env(*args, **kwargs):
        return mock_instance

    mock_instance = MockedDocker()
    monkeypatch.setattr(docker, "from_env", mock_from_env)
    monkeypatch.setattr(docker, "APIClient", MockedAPIClient)
    monkeypatch.setattr(engine.subprocess, "run", mocked_subprocess_run)

    yield mock_instance



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import base64
import json
import subprocess
from collections.abc import Mapping, Sequence
from functools import cached_property
from urllib.parse import urlparse, urlunparse

import numpy as np
import requests
from pydantic import ValidationError
from pydantic_core import InitErrorDetails

from . import engine


class Tesseract:
    """A Tesseract.

    This class represents a single Tesseract instance, either remote or local,
    and provides methods to run commands on it and retrieve results.

    Communication between a Tesseract and this class is done either via
    HTTP requests or `docker exec` invocations (only possible for local
    instances spawned when instantiating the class).
    """

    image: str
    volumes: list[str] | None
    gpus: list[str] | None

    _client: HTTPClient
    project_id: str | None = None
    container_id: str | None = None

    def __init__(self, url: str) -> None:
        self._client = HTTPClient(url)

    @classmethod
    def from_image(
        cls,
        image: str,
        *,
        volumes: list[str] | None = None,
        gpus: list[str] | None = None,
    ):
        obj = cls.__new__(cls)

        obj.image = image
        obj.volumes = volumes
        obj.gpus = gpus

        return obj

    def __enter__(self):
        url = self._serve(volumes=self.volumes, gpus=self.gpus)
        self._client = HTTPClient(url)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        engine.teardown(self.project_id)
        self.project_id = None
        self.container_id = None

    def _serve(
        self,
        port: str = "",
        volumes: list[str] | None = None,
        gpus: list[str] | None = None,
    ) -> str:
        if self.container_id:
            raise RuntimeError(
                "Client already attached to the Tesseract container {self.container_id}"
            )
        project_id = engine.serve([self.image], port=port, volumes=volumes, gpus=gpus)

        command = ["docker", "compose", "-p", project_id, "ps", "--format", "json"]
        result = subprocess.run(command, capture_output=True, text=True)

        # This relies on the fact that result.stdout from docker compose ps
        # contains multiple json dicts, one for each container, separated by newlines,
        # but json.loads will only parse the first one.
        # The first_container dict contains useful info like container id, ports, etc.
        first_container = json.loads(result.stdout)

        if first_container:
            first_container_id = first_container["ID"]
            first_container_port = first_container["Publishers"][0]["PublishedPort"]
        else:
            raise RuntimeError("No containers found.")

        self.project_id = project_id
        self.container_id = first_container_id
        return f"http://localhost:{first_container_port}"

    @cached_property
    def openapi_schema(self) -> dict:
        """Get the OpenAPI schema of this Tessseract.

        Returns:
            dictionary with the OpenAPI Schema.
        """
        return self._client._run_tesseract("openapi_schema")

    @cached_property
    def input_schema(self) -> dict:
        """Get the input schema of this Tessseract.

        Returns:
             dictionary with the input schema.
        """
        return self._client._run_tesseract("input_schema")

    @cached_property
    def output_schema(self) -> dict:
        """Get the output schema of this Tessseract.

        Returns:
             dictionary with the output schema.
        """
        return self._client._run_tesseract("output_schema")

    @property
    def available_endpoints(self) -> list[str]:
        """Get the list of available endpoints.

        Returns:
            a list with all available endpoints for this Tesseract.
        """
        return [endpoint.lstrip("/") for endpoint in self.openapi_schema["paths"]]

    def _run_tesseract(self, endpoint: str, payload: dict | None = None) -> dict:
        return self._client._run_tesseract(endpoint, payload)

    def apply(self, inputs: dict) -> dict:
        """Run apply endpoint.

        Args:
            inputs: a dictionary with the inputs.

        Returns:
            dictionary with the results.
        """
        payload = {"inputs": inputs}
        return self._run_tesseract("apply", payload)

    def abstract_eval(self, inputs: dict) -> dict:
        """Run abstract eval endpoint.

        Args:
            inputs: a dictionary with the (abstract) inputs.

        Returns:
            dictionary with the results.
        """
        payload = {"inputs": inputs}
        return self._run_tesseract("abstract_eval", payload)

    def jacobian(
        self, inputs: dict, jac_inputs: list[str], jac_outputs: list[str]
    ) -> dict:
        """Calculate the Jacobian of (some of the) outputs w.r.t. (some of the) inputs.

        Args:
            inputs: a dictionary with the inputs.
            jac_inputs: Inputs with respect to which derivatives will be calculated.
            jac_outputs: Outputs which will be differentiated.

        Returns:
            dictionary with the results.
        """
        if "jacobian" not in self.available_endpoints:
            raise NotImplementedError("Jacobian not implemented for this Tesseract.")

        payload = {
            "inputs": inputs,
            "jac_inputs": jac_inputs,
            "jac_outputs": jac_outputs,
        }
        return self._run_tesseract("jacobian", payload)

    def jacobian_vector_product(
        self,
        inputs: dict,
        jvp_inputs: list[str],
        jvp_outputs: list[str],
        tangent_vector: dict,
    ) -> dict:
        """Calculate the Jacobian Vector Product (JVP) of (some of the) outputs w.r.t. (some of the) inputs.

        Args:
            inputs: a dictionary with the inputs.
            jvp_inputs: Inputs with respect to which derivatives will be calculated.
            jvp_outputs: Outputs which will be differentiated.
            tangent_vector: Element of the tangent space to multiply with the Jacobian.

        Returns:
            dictionary with the results.
        """
        if "jacobian_vector_product" not in self.available_endpoints:
            raise NotImplementedError(
                "Jacobian Vector Product (JVP) not implemented for this Tesseract."
            )

        payload = {
            "inputs": inputs,
            "jvp_inputs": jvp_inputs,
            "jvp_outputs": jvp_outputs,
            "tangent_vector": tangent_vector,
        }
        return self._run_tesseract("jacobian_vector_product", payload)

    def vector_jacobian_product(
        self,
        inputs: dict,
        vjp_inputs: list[str],
        vjp_outputs: list[str],
        cotangent_vector: dict,
    ) -> dict:
        """Calculate the Vector Jacobian Product (VJP) of (some of the) outputs w.r.t. (some of the) inputs.

        Args:
            inputs: a dictionary with the inputs.
            vjp_inputs: Inputs with respect to which derivatives will be calculated.
            vjp_outputs: Outputs which will be differentiated.
            cotangent_vector: Element of the cotangent space to multiply with the Jacobian.


        Returns:
            dictionary with the results.
        """
        if "vector_jacobian_product" not in self.available_endpoints:
            raise NotImplementedError(
                "Vector Jacobian Product (VJP) not implemented for this Tesseract."
            )

        payload = {
            "inputs": inputs,
            "vjp_inputs": vjp_inputs,
            "vjp_outputs": vjp_outputs,
            "cotangent_vector": cotangent_vector,
        }
        return self._run_tesseract("vector_jacobian_product", payload)


def _tree_map(func, tree, is_leaf=None):
    if is_leaf is not None and is_leaf(tree):
        return func(tree)
    if isinstance(tree, Mapping):  # Dictionary-like structure
        return {key: _tree_map(func, value, is_leaf) for key, value in tree.items()}

    if isinstance(tree, Sequence) and not isinstance(
        tree, (str, bytes)
    ):  # List, tuple, etc.
        return type(tree)(_tree_map(func, item, is_leaf) for item in tree)

    # If nothing above matched do nothing
    return tree


def _encode_array(arr: np.ndarray, b64: bool = True):
    if b64:
        data = {
            "buffer": base64.b64encode(arr.tobytes()).decode(),
            "encoding": "base64",
        }
    else:
        data = {
            "buffer": arr.tolist(),
            "encoding": "raw",
        }

    return {
        "shape": arr.shape,
        "dtype": arr.dtype.name,
        "data": data,
    }


def _decode_array(encoded_arr: dict):
    if "data" in encoded_arr:
        if encoded_arr["data"]["encoding"] == "base64":
            data = base64.b64decode(encoded_arr["data"]["buffer"])
            arr = np.frombuffer(data, dtype=encoded_arr["dtype"])
        else:
            arr = np.array(encoded_arr["data"]["buffer"])
    else:
        arr = encoded_arr
    return arr


class HTTPClient:
    """HTTP Client for Tesseracts."""

    def __init__(self, url: str) -> None:
        self._url = self._sanitize_url(url)

    @staticmethod
    def _sanitize_url(url: str) -> str:
        parsed = urlparse(url)

        if not parsed.scheme:
            url = f"http://{url}"
            parsed = urlparse(url)

        sanitized = urlunparse((parsed.scheme, parsed.netloc, parsed.path, "", "", ""))
        return sanitized

    @property
    def url(self) -> str:
        """(Sanitized) URL to connect to."""
        return self._url

    def _request(
        self, endpoint: str, method: str = "GET", payload: dict | None = None
    ) -> dict:
        url = f"{self.url}/{endpoint.lstrip('/')}"

        if payload:
            encoded_payload = _tree_map(
                _encode_array, payload, is_leaf=lambda x: hasattr(x, "shape")
            )
        else:
            encoded_payload = None

        response = requests.request(method=method, url=url, json=encoded_payload)
        data = response.json()

        if (
            response.status_code == requests.codes.unprocessable_entity
            and "detail" in data
        ):
            errors = [
                InitErrorDetails(
                    type=e["type"],
                    loc=tuple(e["loc"]),
                    input=e.get("input"),
                )
                for e in data["detail"]
            ]
            raise ValidationError.from_exception_data(f"endpoint {endpoint}", errors)
        else:
            response.raise_for_status()

        if endpoint in [
            "apply",
            "jacobian",
            "jacobian_vector_product",
            "vector_jacobian_product",
        ]:
            data = _tree_map(
                _decode_array,
                response.json(),
                is_leaf=lambda x: type(x) is dict and "shape" in x,
            )

        return data

    def _run_tesseract(self, endpoint: str, payload: dict | None = None) -> dict:
        if endpoint in [
            "input_schema",
            "output_schema",
            "openapi_schema",
            "health",
        ]:
            method = "GET"
        else:
            method = "POST"

        if endpoint == "openapi_schema":
            endpoint = "openapi.json"

        return self._request(endpoint, method, payload)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import pytest
from common import build_tesseract, image_exists

from tesseract_core import Tesseract


@pytest.fixture(scope="module")
def built_image_name(docker_client, shared_dummy_image_name, dummy_tesseract_location):
    """Build the dummy Tesseract image for the tests."""
    image_name = build_tesseract(dummy_tesseract_location, shared_dummy_image_name)
    assert image_exists(docker_client, image_name)
    yield image_name


def test_available_endpoints(built_image_name):
    with Tesseract.from_image(built_image_name) as vecadd:
        assert set(vecadd.available_endpoints) == {
            "apply",
            "jacobian",
            "health",
            "input_schema",
            "output_schema",
            "abstract_eval",
            "jacobian_vector_product",
            "vector_jacobian_product",
        }


def test_apply(built_image_name):
    input = {"a": [1, 2], "b": [3, 4], "s": 1}

    with Tesseract.from_image(built_image_name) as vecadd:
        out = vecadd.apply(input)

    np.testing.assert_array_equal(out["result"], np.array([4.0, 6.0]))
    assert set(out.keys()) == {"result"}



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""CLI tests that do not require a running Docker daemon.

(Those go in endtoend_tests/test_endtoend.py.)
"""

import pytest
from typer.testing import CliRunner

from tesseract_core.sdk.cli import app as cli


@pytest.fixture
def cli_runner():
    return CliRunner(mix_stderr=False)


def test_suggestion_on_misspelled_command(cli_runner):
    result = cli_runner.invoke(cli, ["innit"], catch_exceptions=False)
    assert result.exit_code == 2, result.stdout
    assert "No such command 'innit'." in result.stderr
    assert "Did you mean 'init'?" in result.stderr

    result = cli_runner.invoke(cli, ["wellbloodygreatinnit"], catch_exceptions=False)
    assert result.exit_code == 2, result.stdout
    assert "No such command 'wellbloodygreatinnit'." in result.stderr
    assert "Did you mean" not in result.stderr


def test_version(cli_runner):
    from tesseract_core import __version__

    result = cli_runner.invoke(cli, ["--version"])
    assert result.exit_code == 0, result.stdout
    assert __version__ in result.stdout



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Engine to power Tesseract commands."""

import contextlib
import datetime
import json
import linecache
import logging
import optparse
import os
import random
import shlex
import string
import subprocess
import tempfile
import threading
from collections.abc import Callable, Sequence
from pathlib import Path
from shutil import copy, copytree, rmtree
from typing import Any

import docker
import docker.errors
import docker.models
import docker.models.containers
import docker.models.images
from docker.types import DeviceRequest
from jinja2 import Environment, PackageLoader, StrictUndefined
from pip._internal.index.package_finder import PackageFinder
from pip._internal.network.session import PipSession
from pip._internal.req.req_file import (
    RequirementsFileParser,
    get_line_parser,
    handle_line,
)

from .api_parse import TesseractConfig, get_config, validate_tesseract_api
from .exceptions import UserError

logger = logging.getLogger("tesseract")

# Jinja2 Environment
ENV = Environment(
    loader=PackageLoader("tesseract_core.sdk", "templates"),
    undefined=StrictUndefined,
)


class LogPipe(threading.Thread):
    """Custom wrapper to support live logging from a subprocess via a pipe.

    Runs a thread that logs everything read from the pipe to the standard logger.
    Can be used as a context manager for automatic cleanup.
    """

    daemon = True

    def __init__(self, level: int) -> None:
        """Initialize the LogPipe with the given logging level."""
        super().__init__()
        self._level = level
        self._fd_read, self._fd_write = os.pipe()
        self._pipe_reader = os.fdopen(self._fd_read)
        self._captured_lines = []

    def __enter__(self) -> int:
        """Start the thread and return the write file descriptor of the pipe."""
        self.start()
        return self.fileno()

    def __exit__(self, *args: Any) -> None:
        """Close the pipe and join the thread."""
        os.close(self._fd_write)
        # Use a timeout so something weird happening in the logging thread doesn't
        # cause this to hang indefinitely
        self.join(timeout=10)
        # Do not close reader before thread is joined since there may be pending data
        # This also closes the fd_read pipe
        self._pipe_reader.close()

    def fileno(self) -> int:
        """Return the write file descriptor of the pipe."""
        return self._fd_write

    def run(self) -> None:
        """Run the thread, logging everything."""
        for line in iter(self._pipe_reader.readline, ""):
            if line.endswith("\n"):
                line = line[:-1]
            self._captured_lines.append(line)
            logger.log(self._level, line)

    @property
    def captured_lines(self) -> list[str]:
        """Return all lines captured so far."""
        return self._captured_lines


def needs_docker(func: Callable) -> Callable:
    """A decorator for functions that rely on docker daemon."""
    import functools

    @functools.wraps(func)
    def wrapper_needs_docker(*args: Any, **kwargs: Any) -> None:
        try:
            docker.from_env().info()
        except (
            FileNotFoundError,
            docker.errors.APIError,
            docker.errors.DockerException,
        ) as ex:
            raise UserError(
                "Could not reach Docker daemon, check if it is running."
            ) from ex

        return func(*args, **kwargs)

    return wrapper_needs_docker


def parse_requirements(
    filename: str | Path,
    session: PipSession | None = None,
    finder: PackageFinder | None = None,
    options: optparse.Values | None = None,
    constraint: bool = False,
) -> tuple[list[str], list[str]]:
    """Split local dependencies from remote ones in a pip-style requirements file.

    All CLI options that may be part of the given requiremets file are included in
    the remote dependencies.
    """
    if session is None:
        session = PipSession()

    local_dependencies = []
    remote_dependencies = []

    line_parser = get_line_parser(finder)
    parser = RequirementsFileParser(session, line_parser)

    for parsed_line in parser.parse(str(filename), constraint):
        line = linecache.getline(parsed_line.filename, parsed_line.lineno)
        line = line.strip()
        parsed_req = handle_line(
            parsed_line, options=options, finder=finder, session=session
        )
        if not hasattr(parsed_req, "requirement"):
            # this is probably a cli option like --extra-index-url, so we make
            # sure to keep it.
            remote_dependencies.append(line)
        elif parsed_line.requirement.startswith((".", "/", "file://")):
            local_dependencies.append(line)
        else:
            remote_dependencies.append(line)
    return local_dependencies, remote_dependencies


def docker_buildx(
    path: str | Path,
    tag: str,
    dockerfile: str | Path,
    inject_ssh: bool = False,
    keep_build_cache: bool = False,
    print_and_exit: bool = False,
) -> docker.models.images.Image | None:
    """Build a Docker image from a Dockerfile using BuildKit."""
    # Build the Docker image
    # docker-py does not support BuildKit, so we shell out to the Docker CLI
    # see https://github.com/docker/docker-py/issues/2230
    build_cmd = [
        "docker",
        "buildx",
        "build",
        "--load",
        "--tag",
        tag,
        "--file",
        str(dockerfile),
        str(path),
    ]

    if inject_ssh:
        ssh_sock = os.environ.get("SSH_AUTH_SOCK")
        if ssh_sock is None:
            raise ValueError(
                "SSH_AUTH_SOCK environment variable not set (try running `ssh-agent`)"
            )

        ssh_keys = subprocess.run(["ssh-add", "-L"], capture_output=True)
        if ssh_keys.returncode != 0 or not ssh_keys.stdout:
            raise ValueError("No SSH keys found in SSH agent (try running `ssh-add`)")

        build_cmd += ["--ssh", f"default={ssh_sock}"]

    if print_and_exit:
        logger.info(
            f"To build the Docker image manually, run:\n    $ {shlex.join(build_cmd)}"
        )
        return None

    # Record start time for cache pruning -- this isn't perfect, but should be good enough
    # (might prune too much if multiple builds are running at the same time, but that's fine)
    start = datetime.datetime.now()

    out_pipe = LogPipe(logging.DEBUG)
    with out_pipe as out_pipe_fd:
        proc = subprocess.run(build_cmd, stdout=out_pipe_fd, stderr=out_pipe_fd)

    logs = out_pipe.captured_lines
    return_code = proc.returncode

    # NOTE: Do this before error checking to ensure we always prune the cache
    if not keep_build_cache:
        try:
            with contextlib.closing(docker.APIClient()) as api_client:
                api_client.prune_builds(all=True, filters={"until": start.isoformat()})
        except docker.errors.DockerException:
            logger.warning(
                "Docker build cache could not be cleared; consider doing so manually."
            )

    if return_code != 0:
        raise docker.errors.BuildError("Error while building Docker image", logs)

    # Get image object
    with contextlib.closing(docker.from_env()) as client:
        image = client.images.get(tag)

    return image


def get_runtime_dir() -> Path:
    """Get the source directory for the Tesseract runtime as a context manager."""
    import tesseract_core

    return Path(tesseract_core.__file__).parent / "runtime"


def create_dockerfile(user_config: TesseractConfig, use_ssh_mount: bool = False) -> str:
    """Create the Dockerfile for the package.

    Args:
        user_config: The Tesseract configuration object.
        use_ssh_mount: Whether to use SSH mount to install dependencies (prevents caching).

    Returns:
        A string with the Dockerfile content.
    """
    template_name = "Dockerfile.base"
    template = ENV.get_template(template_name)

    template_values = {
        "tesseract_source_directory": "__tesseract_source__",
        "tesseract_runtime_location": "__tesseract_runtime__",
        "config": user_config,
        "use_ssh_mount": use_ssh_mount,
    }

    logger.debug(f"Generating Dockerfile from template: {template_name}")

    return template.render(template_values)


def build_image(
    src_dir: str | Path,
    image_name: str,
    dockerfile: str | Path,
    build_dir: str | Path,
    inject_ssh: bool = False,
    keep_build_cache: bool = False,
    generate_only: bool = False,
) -> docker.models.images.Image | None:
    """Build the image from a Dockerfile.

    Returns:
        An Image object representing an image on the local machine.
    """
    dockerfile_path = build_dir / "Dockerfile"
    with open(dockerfile_path, "w") as f:
        logger.debug(f"Writing Dockerfile to {dockerfile_path}")
        f.write(dockerfile)

    copytree(src_dir, build_dir / "__tesseract_source__")

    if (Path(src_dir) / "tesseract_requirements.txt").exists():
        local_dependencies, remote_dependencies = parse_requirements(
            Path(src_dir) / "tesseract_requirements.txt"
        )
    else:
        remote_dependencies = []
        local_dependencies = []

    if local_dependencies:
        local_requirements_path = build_dir / "local_requirements"
        Path.mkdir(local_requirements_path)
        for dependency in local_dependencies:
            src = Path(src_dir) / dependency
            dest = build_dir / "local_requirements" / src.name
            if src.is_file():
                copy(src, dest)
            else:
                copytree(src, dest)

    # We need to write a new requirements file in the build dir, where we explicitly
    # removed the local dependencies
    requirements_file_path = (
        Path(build_dir) / "__tesseract_source__" / "tesseract_requirements.txt"
    )
    with requirements_file_path.open("w", encoding="utf-8") as f:
        for dependency in remote_dependencies:
            f.write(f"{dependency}\n")

    def _ignore_pycache(_, names: list[str]) -> list[str]:
        ignore = []
        if "__pycache__" in names:
            ignore.append("__pycache__")
        return ignore

    runtime_source_dir = get_runtime_dir()
    copytree(
        runtime_source_dir,
        build_dir / "__tesseract_runtime__" / "tesseract_core" / "runtime",
        ignore=_ignore_pycache,
    )
    for metafile in (runtime_source_dir / "meta").glob("*"):
        copy(metafile, build_dir / "__tesseract_runtime__")

    if generate_only:
        logger.info(f"Build directory generated at {build_dir}, skipping build")
    else:
        logger.info("Building image ...")

    try:
        image = docker_buildx(
            path=build_dir.as_posix(),
            tag=image_name,
            dockerfile=dockerfile_path,
            inject_ssh=inject_ssh,
            keep_build_cache=keep_build_cache,
            print_and_exit=generate_only,
        )

    except docker.errors.BuildError as e:
        logger.warning("Build failed with logs:")
        for line in e.build_log:
            logger.warning(line)
        raise e
    else:
        if image is not None:
            logger.debug("Build successful")

    return image


def _write_template_file(
    template_name: str,
    target_dir: Path,
    template_vars: dict,
    recipe: Path = Path("."),
    exist_ok: bool = False,
):
    """Write a template to a target directory."""
    template = ENV.get_template(str(recipe / template_name))

    target_file = target_dir / template_name

    if target_file.exists() and not exist_ok:
        raise FileExistsError(f"File {target_file} already exists")

    logger.info(f"Writing template {template_name} to {target_file}")

    with open(target_file, "w") as target_fp:
        target_fp.write(template.render(template_vars))

    return target_file


def init_api(
    target_dir: Path,
    tesseract_name: str,
    recipe: str = "base",
) -> Path:
    """Create a new empty Tesseract API module at the target location."""
    from tesseract_core import __version__ as tesseract_version

    template_vars = {
        "version": tesseract_version,
        "timestamp": datetime.datetime.now().isoformat(),
        "name": tesseract_name,
    }

    # If target dir does not exist, create it
    Path(target_dir).mkdir(parents=True, exist_ok=True)

    _write_template_file(
        "tesseract_api.py", target_dir, template_vars, recipe=Path(recipe)
    )
    _write_template_file(
        "tesseract_config.yaml", target_dir, template_vars, recipe=Path(recipe)
    )
    _write_template_file(
        "tesseract_requirements.txt", target_dir, template_vars, recipe=Path(recipe)
    )

    return target_dir / "tesseract_api.py"


def build_tesseract(
    src_dir: Path,
    image_tag: str | None,
    build_dir: Path | None = None,
    inject_ssh: bool = False,
    config_override: tuple[tuple[list[str], str], ...] = (),
    keep_build_cache: bool = False,
    generate_only: bool = False,
) -> docker.models.images.Image | Path:
    """Build a new Tesseract from a context directory.

    Args:
        src_dir: path to the Tesseract project directory, where the
          `tesseract_api.py` and `tesseract_config.yaml` files
          are located.
        image_tag: name to be used as a tag for the Tesseract image.
        build_dir: directory to be used to store the build context.
          If not provided, a temporary directory will be created.
        inject_ssh: whether or not to forward SSH agent when building the image.
        config_override: overrides for configuration options in the Tesseract.
        keep_build_cache: whether or not to keep the Docker build cache.
        generate_only: only generate the build context but do not build the image.

    Returns:
        docker.models.images.Image representing the built Tesseract image,
        or path to build directory if `generate_only` is True.
    """
    validate_tesseract_api(src_dir)
    config = get_config(src_dir)

    # Apply config overrides
    for path, value in config_override:
        c = config
        for k in path[:-1]:
            c = getattr(c, k)
        setattr(c, path[-1], value)

    image_name = config.name
    if image_tag:
        image_name += f":{image_tag}"

    source_basename = Path(src_dir).name

    if build_dir is None:
        build_dir = Path(tempfile.mkdtemp(prefix=f"tesseract_build_{source_basename}"))
        keep_build_dir = True if generate_only else False
    else:
        build_dir = Path(build_dir)
        build_dir.mkdir(exist_ok=True)
        keep_build_dir = True

    dockerfile = create_dockerfile(config, use_ssh_mount=inject_ssh)

    try:
        out = build_image(
            src_dir,
            image_name,
            dockerfile,
            build_dir=build_dir,
            inject_ssh=inject_ssh,
            keep_build_cache=keep_build_cache,
            generate_only=generate_only,
        )
    finally:
        if not keep_build_dir:
            try:
                rmtree(build_dir)
            except OSError as exc:
                # Permission denied or already removed
                logger.info(
                    f"Could not remove temporary build directory {build_dir}: {exc}"
                )
                pass

    if generate_only:
        return build_dir

    return out


def teardown(project_id: str) -> None:
    """Teardown Tesseract image(s) running in a Docker Compose project.

    Args:
        project_id: Docker Compose project ID to teardown.
    """
    if not project_id:
        raise ValueError("Docker Compose project ID is empty or None, cannot teardown")
    if not _docker_compose_project_exists(project_id):
        raise ValueError(
            f"A Docker Compose project with ID {project_id} cannot be found, use `docker compose ls` to find project ID"
        )

    if not _docker_compose_down(project_id):
        raise RuntimeError(
            f"Cannot teardown Docker Compose project with ID: {project_id}"
        )


def get_tesseract_containers() -> list[docker.models.containers.Container]:
    """Get Tesseract containers."""
    return list(
        filter(
            lambda container: _is_tesseract(container),
            docker.from_env().containers.list(),
        )
    )


def get_tesseract_images() -> list[docker.models.images.Image]:
    """Get Tesseract images."""
    return list(filter(lambda img: _is_tesseract(img), docker.from_env().images.list()))


def _get_docker_image(image_name: str) -> docker.models.images.Image:
    """Get Docker image object."""
    try:
        return docker.from_env().images.get(image_name)
    except docker.errors.ImageNotFound:
        logger.error(f"No Docker image found with name: {image_name}")
        return None


def _is_tesseract(
    docker_asset: docker.models.images.Image | docker.models.containers.Container,
) -> bool:
    """Check if an image is Tesseract."""
    if not any(
        "TESSERACT_NAME" in env_val for env_val in docker_asset.attrs["Config"]["Env"]
    ):
        return False
    return True


def serve(
    images: list[str | docker.models.images.Image],
    port: str = "",
    volumes: list[str] | None = None,
    gpus: list[str] | None = None,
) -> str:
    """Serve one or more Tesseract images.

    Start the Tesseracts listening on an available ports on the host.

    Args:
        images: a list of Tesseract image IDs as strings or `docker`'s
                Image object.
        port: port or port range to serve the tesseract on.
        volumes: list of paths to mount in the Tesseract container.
        gpus: IDs of host Nvidia GPUs to make available to the Tesseracts.

    Returns:
        A string representing the Tesseract Project ID.
    """
    if not images or not all(
        (isinstance(item, str) or isinstance(item, docker.models.images.Image))
        for item in images
    ):
        raise ValueError("One or more Tesseract image IDs must be provided")

    image_ids = []
    for image_ in images:
        if isinstance(image_, docker.models.images.Image):
            image = image_
        else:  # str
            image = _get_docker_image(image_)

        if not image:
            raise ValueError(f"Image ID {image_} is not a valid Docker image")
        if not _is_tesseract(image):
            raise ValueError(f"Input ID {image.id} is not a valid Tesseract")
        image_ids.append(image.id)

    template = _create_docker_compose_template(image_ids, port, volumes, gpus)
    compose_fname = _create_compose_fname()

    with tempfile.NamedTemporaryFile(
        mode="w+",
        prefix=compose_fname,
    ) as compose_file:
        compose_file.write(template)
        compose_file.flush()

        project_name = _create_compose_proj_id()
        if not _docker_compose_up(compose_file.name, project_name):
            raise RuntimeError("Cannot serve Tesseracts")
        return project_name


def _docker_compose_project_exists(project_id: str) -> bool:
    """Check if Docker Compose project exists."""
    try:
        result = subprocess.run(
            ["docker", "compose", "ls", "-a", "--format", "json"],
            capture_output=True,
            text=True,
            check=True,
        )
        if not any(
            project["Name"] == project_id for project in json.loads(result.stdout)
        ):
            logger.error(f"Docker Compose project with ID {project_id} does not exist")
            return False
        return True
    except (subprocess.CalledProcessError, json.JSONDecodeError) as ex:
        logger.error(str(ex))
        return False


def _docker_compose_down(project_id: str) -> bool:
    """Stop and remove containers and networks associated to a project."""
    try:
        __ = subprocess.run(
            ["docker", "compose", "-p", project_id, "down"],
            check=True,
            capture_output=True,
        )
        return True
    except subprocess.CalledProcessError as ex:
        logger.error(str(ex))
        return False


def _docker_compose_up(compose_fpath: str, project_name: str) -> bool:
    """Start containers using Docker Compose template."""
    logger.info("Waiting for Tesseract containers to start ...")
    try:
        _ = subprocess.run(
            [
                "docker",
                "compose",
                "-f",
                compose_fpath,
                "-p",
                project_name,
                "up",
                "-d",
                "--wait",
            ],
            check=True,
            capture_output=True,
        )
        return True
    except subprocess.CalledProcessError as ex:
        logger.error(str(ex))
        logger.error(ex.stderr.decode())
        return False


def _create_docker_compose_template(
    image_ids: list[str],
    port: str = "",
    volumes: list[str] | None = None,
    gpus: list[str] | None = None,
) -> str:
    """Create Docker Compose template."""
    services = []
    if not port:
        port = "8000"
    else:
        port = f"{port}:8000"

    gpu_settings = None
    if gpus:
        if (len(gpus) == 1) and (gpus[0] == "all"):
            gpu_settings = "count: all"
        else:
            gpu_settings = f"device_ids: {gpus}"

    for image_id in image_ids:
        service = {
            "name": _create_compose_service_id(image_id),
            "image": image_id,
            "port": port,
            "volumes": volumes,
            "gpus": gpu_settings,
        }

        services.append(service)
    template = ENV.get_template("docker-compose.yml")
    return template.render(services=services)


def _create_compose_service_id(image_id: str) -> str:
    """Create Docker Compose service ID."""
    image_id = image_id.split(":")[0]
    return f"{image_id}-{_id_generator()}"


def _create_compose_proj_id() -> str:
    """Create Docker Compose project ID."""
    return f"tesseract-{_id_generator()}"


def _create_compose_fname() -> str:
    """Create Docker Compose project file name."""
    return f"docker-compose-{_id_generator()}.yml"


def _id_generator(
    size: int = 12, chars: Sequence[str] = string.ascii_lowercase + string.digits
) -> str:
    """Generate ID."""
    return "".join(random.choice(chars) for _ in range(size))


def _parse_volumes(options: list[str]) -> dict[str, dict[str, str]]:
    """Parses volume mount strings to dict accepted by docker SDK.

    Strings of the form 'source:target:(ro|rw)' are parsed to
    `{source: {'bind': target, 'mode': '(ro|rw)'}}`.
    """

    def _parse_option(option: str):
        args = option.split(":")
        if len(args) == 2:
            source, target = args
            mode = "ro"
        elif len(args) == 3:
            source, target, mode = args
        else:
            raise ValueError(
                f"Invalid mount volume specification {option} "
                "(must be `/path/to/source:/path/totarget:(ro|rw)`)",
            )
        # Docker doesn't like paths like ".", so we convert to absolute path here
        source = str(Path(source).resolve())
        return source, {"bind": target, "mode": mode}

    return dict(_parse_option(opt) for opt in options)


def run_tesseract(
    image: str | docker.models.images.Image,
    command: str,
    args: list[str],
    volumes: list[str] | None = None,
    gpus: list[int | str] | None = None,
) -> tuple[str, str]:
    """Start a Tesseract and execute a given command.

    Args:
        image: string or docker.models.images.Image object of the Tesseract to run.
        command: Tesseract command to run, e.g. apply.
        args: arguments for the command.
        volumes: list of paths to mount in the Tesseract container.
        gpus: list of GPUs, as indices or names, to passthrough the container.

    Returns:
        Tuple with the stdout and stderr of the Tesseract.
    """
    client = docker.from_env()

    # Args that require rw access to the mounted volume
    output_args = {"-o", "--output-path"}

    cmd = [command]
    current_cmd = None

    if volumes is None:
        parsed_volumes = {}
    else:
        parsed_volumes = _parse_volumes(volumes)

    if gpus is None:
        device_requests = None
    else:
        device_requests = [DeviceRequest(device_ids=gpus, capabilities=[["gpu"]])]

    for arg in args:
        if arg.startswith("-"):
            current_cmd = arg
            cmd.append(arg)
            continue

        # Mount local output directories into Docker container as a volume
        if current_cmd in output_args and "://" not in arg:
            if arg.startswith("@"):
                raise ValueError(
                    f"Output path {arg} cannot start with '@' (used only for input files)"
                )

            local_path = Path(arg).resolve()
            local_path.mkdir(parents=True, exist_ok=True)

            if not local_path.is_dir():
                raise RuntimeError(
                    f"Path {local_path} provided as output is not a directory"
                )

            path_in_container = "/mnt/output"
            arg = path_in_container

            # Bind-mount directory
            parsed_volumes[str(local_path)] = {"bind": path_in_container, "mode": "rw"}

        # Mount local input files marked by @ into Docker container as a volume
        elif arg.startswith("@") and "://" not in arg:
            local_path = Path(arg.lstrip("@")).resolve()

            if not local_path.is_file():
                raise RuntimeError(f"Path {local_path} provided as input is not a file")

            path_in_container = os.path.join("/mnt", f"payload{local_path.suffix}")
            arg = f"@{path_in_container}"

            # Bind-mount file
            parsed_volumes[str(local_path)] = {"bind": path_in_container, "mode": "ro"}

        current_cmd = None
        cmd.append(arg)

    # Run the container
    if isinstance(image, docker.models.images.Image):
        image_id = image.short_id
    else:
        image_id = image
    container = None
    try:
        container = client.containers.run(
            image=image_id,
            command=cmd,
            volumes=parsed_volumes,
            detach=True,
            device_requests=device_requests,
        )
        result = container.wait()
        stdout = container.logs(stdout=True, stderr=False)
        stderr = container.logs(stdout=False, stderr=True)

        exit_code = result["StatusCode"]
        if exit_code != 0:
            stderr = f"\n{stderr.decode('utf-8')}"
            raise docker.errors.ContainerError(
                container, exit_code, shlex.join(cmd), image_id, stderr
            )
    finally:
        if container is not None:
            container.remove(v=True, force=True)

    return stdout.decode("utf-8"), stderr.decode("utf-8")


def exec_tesseract(
    container_id: str,
    command: str,
    args: list[str],
) -> tuple[str, str]:
    """Execute a given command on an existing Tesseract container.

    See `run_tesseract` for the equivalent command that operates on Tesseract images.

    Args:
        container_id: id of the target Tesseract container.
        command: Tesseract command to run, e.g. apply.
        args: arguments for the command.

    Returns:
        Tuple with the stdout and stderr of the Tesseract.
    """
    try:
        result = subprocess.run(
            ["docker", "exec", container_id, "tesseract-runtime", command, *args],
            capture_output=True,
            text=True,
            check=True,
        )
        return result.stdout, result.stderr
    except subprocess.CalledProcessError as e:
        return e.stdout, e.stderr


def project_containers(
    project_id: str,
) -> list[docker.models.containers.Container]:
    """Find containers associated with a Docker Compose Project ID.

    Args:
        project_id: the Docker Compose project ID.

    Returns:
        A list of Docker Images.
    """
    client = docker.from_env()
    return list(filter(lambda x: project_id in x.name, client.containers.list()))



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
import random
import time
from pathlib import Path
from unittest.mock import Mock

import docker
import docker.models
import docker.models.containers
import docker.models.images
import pytest
import yaml
from jinja2.exceptions import TemplateNotFound
from typeguard import suppress_type_checks

from tesseract_core.sdk import engine
from tesseract_core.sdk.api_parse import TesseractConfig, validate_tesseract_api
from tesseract_core.sdk.cli import AVAILABLE_RECIPES
from tesseract_core.sdk.exceptions import UserError


def test_create_dockerfile():
    """Test we can create a dockerfile."""
    default_config = TesseractConfig(name="foobar")
    engine.create_dockerfile(default_config)


@pytest.mark.parametrize("generate_only", [True, False])
def test_build_image(
    dummy_tesseract_package, tmpdir, mocked_docker, generate_only, caplog
):
    """Test we can build an image for a package and keep build directory."""
    src_dir = dummy_tesseract_package
    image_name = "foo.bar/baz:42"
    dockerfile = "FROM tesseract"

    with caplog.at_level(logging.INFO):
        got = engine.build_image(
            src_dir=src_dir,
            image_name=image_name,
            dockerfile=dockerfile,
            build_dir=Path(tmpdir),
            generate_only=generate_only,
        )

    # Check if Dockerfile is there
    dockerfile_path = tmpdir / "Dockerfile"
    assert dockerfile_path.exists()

    if generate_only:
        # Check stdout if it contains the correct docker build command
        assert got is None
        command = f"docker buildx build --load --tag {image_name} --file {dockerfile_path} {tmpdir}"
        assert command in caplog.text
    else:
        assert got.attrs == mocked_docker.images.get(image_name).attrs

    with open(dockerfile_path) as got_dockerfile:
        assert got_dockerfile.read() == dockerfile


@pytest.mark.parametrize("recipe", [None, *AVAILABLE_RECIPES])
def test_init(tmpdir, recipe):
    """Test the initialization of a tesseract from the template."""
    if recipe:
        api_path = engine.init_api(
            target_dir=Path(tmpdir) / "test_dir", tesseract_name="foo", recipe=recipe
        )
    else:
        api_path = engine.init_api(
            target_dir=Path(tmpdir) / "test_dir", tesseract_name="foo"
        )

    # Make sure that all tesseract related files are created
    assert api_path.exists()
    assert (tmpdir / "test_dir/tesseract_requirements.txt").exists()
    assert (tmpdir / "test_dir/tesseract_config.yaml").exists()

    # Ensure the name in the config is correct
    with open(tmpdir / "test_dir/tesseract_config.yaml") as config_yaml:
        assert yaml.safe_load(config_yaml)["name"] == "foo"

    # Ensure template passes validation
    validate_tesseract_api(api_path.parent)

    if not recipe:
        # Ensure it still passes when commenting in optional endpoints
        with open(api_path) as f:
            api_code = f.read()

        start_idx = api_code.find("# Optional endpoints")
        assert start_idx != -1

        api_code = api_code[: start_idx + 1] + api_code[start_idx + 1 :].replace(
            "# ", ""
        )

        with open(api_path, "w") as f:
            f.write(api_code)

    validate_tesseract_api(api_path.parent)


def test_init_bad_recipe(tmpdir):
    """Test the initialization of a tesseract with a bad recipe.

    This does not check for pretty terminal output or typer validation.
    But ensures that an error is raised if there are missing template files.
    """
    with pytest.raises(TemplateNotFound):
        engine.init_api(
            target_dir=Path(tmpdir) / "test_dir",
            tesseract_name="foo",
            recipe="recipewillneverexist",
        )


def test_run_tesseract(mocked_docker):
    """Test running a tesseract."""
    res_out, res_err = engine.run_tesseract(
        "foobar", "apply", ['{"inputs": {"a": [1, 2, 3], "b": [4, 5, 6]}}']
    )

    # Mocked docker just returns the kwargs to `docker run` as json
    res = json.loads(res_out)
    assert res["command"] == ["apply", '{"inputs": {"a": [1, 2, 3], "b": [4, 5, 6]}}']
    assert res["image"] == "foobar"

    # Also check that stderr is captured
    assert res_err == "hello tesseract"

    # Check that we did not request GPUs by accident
    assert res["device_requests"] is None


def test_run_gpu(mocked_docker):
    """Test running a tesseract with all available GPUs."""
    res_out, _ = engine.run_tesseract(
        "foobar",
        "apply",
        ['{"inputs": {"a": [1, 2, 3], "b": [4, 5, 6]}}'],
        gpus=["all"],
    )

    res = json.loads(res_out)
    assert res["device_requests"][0]["DeviceIDs"] == ["all"]


def test_run_tesseract_file_input(mocked_docker, tmpdir):
    """Test running a tesseract with file input / output."""
    outdir = Path(tmpdir) / "output"
    outdir.mkdir()

    infile = Path(tmpdir) / "input.json"
    infile.touch()

    res, _ = engine.run_tesseract(
        "foobar",
        "apply",
        [f"@{infile}", "--output-path", str(outdir)],
    )

    # Mocked docker just returns the kwargs to `docker run` as json
    res = json.loads(res)
    assert res["command"] == [
        "apply",
        "@/mnt/payload.json",
        "--output-path",
        "/mnt/output",
    ]
    assert res["image"] == "foobar"
    assert res["volumes"].keys() == {str(infile), str(outdir)}

    # Test the same with a folder mount
    res, _ = engine.run_tesseract(
        "foobar",
        "apply",
        [f"@{infile}", "--output-path", str(outdir)],
        volumes=["/path/on/host:/path/in/container:ro"],
    )
    res = json.loads(res)
    assert res["volumes"].keys() == {str(infile), str(outdir), "/path/on/host"}
    assert res["volumes"]["/path/on/host"] == {
        "mode": "ro",
        "bind": "/path/in/container",
    }


def test_serve_tesseracts_invalid_input_args():
    """Test input validation logic for multi-tesseract serve."""
    with suppress_type_checks():
        with pytest.raises(ValueError):
            engine.serve([None])

        with pytest.raises(ValueError):
            engine.serve([])

        with pytest.raises(ValueError):
            engine.serve([None, "vectoradd"])

        with pytest.raises(ValueError):
            engine.teardown(None)


def test_get_tesseract_images(mocked_docker):
    tesseract_images = engine.get_tesseract_images()
    assert len(tesseract_images) == 1


def test_get_tesseract_containers(mocked_docker):
    tesseract_containers = engine.get_tesseract_containers()
    assert len(tesseract_containers) == 1


def test_serve_tesseracts(mocked_docker):
    """Test multi-tesseract serve."""
    # Serve valid
    project_name_single_tesseract = engine.serve(["vectoradd"])
    assert project_name_single_tesseract

    # Teardown valid
    engine.teardown(project_name_single_tesseract)

    # Multi-serve valid
    project_name_multi_tesseract = engine.serve(["vectoradd", "vectoradd"])
    assert project_name_multi_tesseract

    # Multi-teardown valid
    engine.teardown(project_name_multi_tesseract)

    # Tear down invalid
    with pytest.raises(ValueError):
        engine.teardown("invalid_project_name")

    # Serve with gpus
    project_name_multi_tesseract = engine.serve(["vectoradd"], gpus=["1", "3"])
    assert project_name_multi_tesseract


def test_needs_docker(mocked_docker):
    @engine.needs_docker
    def run_something_with_docker():
        pass

    # Happy case
    run_something_with_docker()

    mocked_docker.info = Mock(side_effect=docker.errors.APIError(""))

    with pytest.raises(UserError):
        run_something_with_docker()


def test_logpipe(caplog):
    # Verify that logging in a separate thread works as intended
    from tesseract_core.sdk.engine import LogPipe

    caplog.set_level(logging.INFO, logger="tesseract")

    logged_lines = []
    for _ in range(100):
        msg_length = 2 ** random.randint(1, 12)
        msg = "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=msg_length))
        logged_lines.append(msg)

    logpipe = LogPipe(logging.INFO)
    with logpipe:
        fd = os.fdopen(logpipe.fileno(), "w", closefd=False)
        for line in logged_lines:
            print(line, file=fd)
            time.sleep(random.random() / 100)
        fd.flush()

    assert logpipe.captured_lines == logged_lines
    assert caplog.record_tuples == [
        ("tesseract", logging.INFO, line) for line in logged_lines
    ]


def test_parse_requirements(tmpdir):
    reqs = """
    --extra-index-url https://download.pytorch.org/whl/cpu
    torch==2.5.1

    --find-links https://data.pyg.org/whl/torch-2.5.1+cpu.html
    torch_scatter==2.1.2+pt25cpu

    ./internal_packages/foobar
    """
    reqs_file = Path(tmpdir) / "requirements.txt"
    with open(reqs_file, "w") as fi:
        fi.write(reqs)
    locals, remotes = engine.parse_requirements(reqs_file)

    assert locals == [
        "./internal_packages/foobar",
    ]
    assert remotes == [
        "--extra-index-url https://download.pytorch.org/whl/cpu",
        "torch==2.5.1",
        "--find-links https://data.pyg.org/whl/torch-2.5.1+cpu.html",
        "torch_scatter==2.1.2+pt25cpu",
    ]



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import os
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict, FilePath


class RuntimeConfig(BaseModel):
    """Available runtime configuration."""

    tesseract_api_path: FilePath = Path("tesseract_api.py")
    tesseract_name: str = "Tesseract"
    tesseract_version: str = "0+unknown"

    model_config = ConfigDict(frozen=True, extra="forbid")


def update_config(**kwargs: Any) -> None:
    """Create a new runtime configuration from the current environment.

    Passed keyword arguments will override environment variables.
    """
    global _current_config

    conf_settings = {}
    for field in RuntimeConfig.model_fields.keys():
        env_key = field.upper()
        if env_key in os.environ:
            conf_settings[field] = os.environ[env_key]

    conf_settings.update(kwargs)

    config = RuntimeConfig(**conf_settings)
    _current_config = config


_current_config = None
update_config()


def get_config() -> RuntimeConfig:
    """Return the current runtime configuration."""
    return _current_config



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""End-to-end tests for Tesseract workflows."""

import json
from pathlib import Path

import pytest
import requests
import yaml
from common import build_tesseract, image_exists
from typer.testing import CliRunner

from tesseract_core.sdk.cli import AVAILABLE_RECIPES, app


@pytest.fixture(scope="module")
def built_image_name(docker_client, shared_dummy_image_name, dummy_tesseract_location):
    """Build the dummy Tesseract image for the tests."""
    image_name = build_tesseract(dummy_tesseract_location, shared_dummy_image_name)
    assert image_exists(docker_client, image_name)
    yield image_name


@pytest.mark.parametrize("tag", [True, False])
@pytest.mark.parametrize("recipe", [None, *AVAILABLE_RECIPES])
def test_build_from_init_endtoend(
    docker_client, dummy_image_name, tmp_path, tag, recipe
):
    """Test that a trivial (empty) Tesseract image can be built from init."""
    cli_runner = CliRunner(mix_stderr=False)

    init_args = ["init", "--target-dir", str(tmp_path), "--name", dummy_image_name]
    if recipe:
        init_args.extend(["--recipe", recipe])

    result = cli_runner.invoke(app, init_args, catch_exceptions=False)
    assert result.exit_code == 0, result.stderr
    assert (tmp_path / "tesseract_api.py").exists()
    with open(tmp_path / "tesseract_config.yaml") as config_yaml:
        assert yaml.safe_load(config_yaml)["name"] == dummy_image_name

    img_tag = "foo" if tag else None
    image_name = build_tesseract(tmp_path, dummy_image_name, tag=img_tag)
    assert image_exists(docker_client, image_name)

    # Test that the image can be run and that --help is forwarded correctly
    result = cli_runner.invoke(
        app,
        [
            "run",
            image_name,
            "apply",
            "--help",
        ],
        catch_exceptions=False,
    )
    assert f"Usage: tesseract run {image_name} apply" in result.stderr


def test_build_generate_only(dummy_tesseract_location):
    """Test output of build with --generate_only flag."""
    cli_runner = CliRunner(mix_stderr=False)
    build_res = cli_runner.invoke(
        app,
        [
            "build",
            str(dummy_tesseract_location),
            "--generate-only",
        ],
        # Ensure that the output is not truncated
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert build_res.exit_code == 0, build_res.stderr
    # Check that stdout contains "docker buildx"
    command = "docker buildx build"
    assert command in build_res.stderr

    build_dir = Path(build_res.stdout.strip())
    assert build_dir.exists()
    dockerfile_path = build_dir / "Dockerfile"
    assert dockerfile_path.exists()


def test_tesseract_list(built_image_name):
    # Test List Command
    cli_runner = CliRunner(mix_stderr=False)

    list_res = cli_runner.invoke(
        app,
        [
            "list",
        ],
        # Ensure that the output is not truncated
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert list_res.exit_code == 0, list_res.stderr
    assert built_image_name.split(":")[0] in list_res.stdout


def test_tesseract_run_stdout(built_image_name):
    # Test List Command
    cli_runner = CliRunner(mix_stderr=False)

    test_commands = ("input-schema", "output-schema", "openapi-schema", "health")

    for command in test_commands:
        run_res = cli_runner.invoke(
            app,
            [
                "run",
                built_image_name,
                command,
            ],
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr
        assert run_res.stdout

        try:
            json.loads(run_res.stdout)
        except json.JSONDecodeError:
            print(f"failed to decode {command} stdout as JSON")
            print(run_res.stdout)
            raise


def test_tesseract_serve_pipeline(docker_client, built_image_name):
    cli_runner = CliRunner(mix_stderr=False)
    try:
        run_res = cli_runner.invoke(
            app,
            [
                "serve",
                built_image_name,
            ],
            catch_exceptions=False,
        )

        assert run_res.exit_code == 0, run_res.stderr
        assert run_res.stdout

        project_meta = json.loads(run_res.stdout)

        project_id = project_meta["project_id"]
        project_containers = [
            c for c in docker_client.containers.list() if project_id in c.name
        ]
        if not project_containers:
            raise ValueError(f"Could not find container for project '{project_id}'")

        project_container = project_containers[0]
        assert project_container.name == project_meta["containers"][0]["name"]

        port_key = next(iter(project_container.ports))
        port = project_container.ports[port_key][0]["HostPort"]
        assert port == project_meta["containers"][0]["port"]

        # Ensure served Tesseract is usable
        res = requests.get(f"http://localhost:{port}/health")
        assert res.status_code == 200, res.text

        # Ensure project id is shown in `tesseract ps`
        run_res = cli_runner.invoke(
            app,
            ["ps"],
            env={"COLUMNS": "1000"},
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr
        assert project_id in run_res.stdout
        assert port in run_res.stdout
        assert project_container.short_id in run_res.stdout
    finally:
        run_res = cli_runner.invoke(
            app,
            [
                "teardown",
                project_id,
            ],
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr


@pytest.mark.parametrize("tear_all", [True, False])
def test_tesseract_teardown_multiple(built_image_name, tear_all):
    """Teardown multiple projects."""
    cli_runner = CliRunner(mix_stderr=False)

    project_ids = []
    try:
        for _ in range(0, 5):
            # Serve
            run_res = cli_runner.invoke(
                app,
                [
                    "serve",
                    built_image_name,
                ],
                catch_exceptions=False,
            )
            assert run_res.exit_code == 0, run_res.stderr
            assert run_res.stdout

            project_meta = json.loads(run_res.stdout)

            project_id = project_meta["project_id"]
            project_ids.append(project_id)

    finally:
        # Teardown multiple/all
        args = ["teardown"]
        if tear_all:
            args.extend(["--all"])
        else:
            args.extend(project_ids)

        run_res = cli_runner.invoke(
            app,
            args,
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr
        # Ensure all projects are killed
        run_res = cli_runner.invoke(
            app,
            ["ps"],
            env={"COLUMNS": "1000"},
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr
        for project_id in project_ids:
            assert project_id not in run_res.stdout


def test_tesseract_serve_ports_error(built_image_name):
    """Check error handling for serve -p flag."""
    cli_runner = CliRunner(mix_stderr=False)

    # Check multiple Tesseracts being served.
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            built_image_name,
            built_image_name,
            "-p",
            "8000-8001",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code
    assert (
        "Port specification only works if 1 Tesseract is being served."
        in run_res.stderr
    )

    # Check invalid ports.
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            "-p",
            "8000-999999",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code
    assert "Ports '8000-999999' must be between 1025 and 65535." in run_res.stderr

    # Check poorly formatted ports.
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            "-p",
            "8000:8081",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code
    assert "Port '8000:8081' must be single integer or a range" in run_res.stderr

    # Check invalid port range.
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            "-p",
            "8000-7000",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code
    assert "Start port '8000' must be less than or equal to end" in run_res.stderr


@pytest.mark.parametrize("port", ["34567", "34567-34569"])
def test_tesseract_serve_ports(built_image_name, port):
    """Try to serve multiple Tesseracts on multiple ports."""
    cli_runner = CliRunner(mix_stderr=False)

    # Serve tesseract on specified ports.
    run_res = cli_runner.invoke(
        app,
        ["serve", built_image_name, "-p", port],
        catch_exceptions=False,
    )
    assert run_res.exit_code == 0, run_res.stderr
    assert run_res.stdout

    project_meta = json.loads(run_res.stdout)
    project_id = project_meta["project_id"]

    # Wrap test in try-finally to ensure teardown of served Tesseract.
    try:
        # Ensure that actual used ports are in the specified port range.
        test_ports = port.split("-")
        start_port = int(test_ports[0])
        end_port = int(test_ports[1]) if len(test_ports) > 1 else start_port

        port = int(project_meta["containers"][0]["port"])
        assert port in range(start_port, end_port + 1)

        # Ensure specified ports are in `tesseract ps` and served Tesseracts are usable.
        run_res = cli_runner.invoke(
            app,
            ["ps"],
            env={"COLUMNS": "1000"},
            catch_exceptions=False,
        )

        res = requests.get(f"http://localhost:{port}/health")
        assert res.status_code == 200, res.text
        assert str(port) in run_res.stdout
    finally:
        run_res = cli_runner.invoke(
            app,
            [
                "teardown",
                project_id,
            ],
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr


def test_tesseract_serve_with_volumes(built_image_name, tmp_path, docker_client):
    """Try to serve multiple Tesseracts with volume mounting."""
    cli_runner = CliRunner(mix_stderr=False)

    # Pytest creates the tmp_path fixture with drwx------ mode, we need others
    # to be able to read and execute the path so the Docker volume is readable
    # from within the container
    tmp_path.chmod(0o0707)

    dest = Path("/foo/")
    try:
        run_res = cli_runner.invoke(
            app,
            [
                "serve",
                "--volume",
                f"{tmp_path}:{dest}",
                built_image_name,
                built_image_name,
            ],
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr
        assert run_res.stdout

        project_meta = json.loads(run_res.stdout)
        project_id = project_meta["project_id"]
        tesseract0_id = project_meta["containers"][0]["name"]
        tesseract0 = docker_client.containers.get(tesseract0_id)
        tesseract1_id = project_meta["containers"][1]["name"]
        tesseract1 = docker_client.containers.get(tesseract1_id)

        # Create file outside the containers and check it from inside the container
        tmpfile = Path(tmp_path) / "hi"
        with open(tmpfile, "w") as hello:
            hello.write("world")
            hello.flush()

        exit_code, output = tesseract0.exec_run(["cat", f"{dest}/hi"])
        assert exit_code == 0
        assert output.decode() == "world"

        # Create file inside a container and check it from the other
        bar_file = dest / "bar"
        exit_code, output = tesseract0.exec_run(["touch", f"{bar_file}"])
        assert exit_code == 0
        exit_code, output = tesseract1.exec_run(["cat", f"{bar_file}"])
        assert exit_code == 0

        # The file should exist outside the container
        assert (tmp_path / "bar").exists()
    finally:
        run_res = cli_runner.invoke(
            app,
            [
                "teardown",
                project_id,
            ],
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr


def test_tesseract_cli_options_parsing(built_image_name, tmpdir):
    cli_runner = CliRunner(mix_stderr=False)

    tmpdir.chmod(0o0707)

    examples_dir = Path(__file__).parent.parent.parent / "examples"
    example_inputs = examples_dir / "vectoradd" / "example_inputs.json"

    test_commands = (
        ["apply", "-f", "json+binref", "-o", str(tmpdir), f"@{example_inputs}"],
        ["apply", f"@{example_inputs}", "-f", "json+binref", "-o", str(tmpdir)],
        ["apply", "-o", str(tmpdir), f"@{example_inputs}", "-f", "json+binref"],
    )

    for args in test_commands:
        run_res = cli_runner.invoke(
            app,
            [
                "run",
                built_image_name,
                *args,
            ],
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr

        with open(Path(tmpdir) / "results.json") as fi:
            results = fi.read()
            assert ".bin:0" in results


def test_tarball_install(dummy_tesseract_package):
    import subprocess
    from textwrap import dedent

    tesseract_api = dedent(
        """
    import cowsay
    from pydantic import BaseModel

    class InputSchema(BaseModel):
        message: str = "Hello, Tesseractor!"

    class OutputSchema(BaseModel):
        out: str

    def apply(inputs: InputSchema) -> OutputSchema:
        return OutputSchema(out=cowsay.get_output_string("cow", inputs.message))
    """
    )

    tesseract_requirements = "./cowsay-6.1-py3-none-any.whl"

    subprocess.run(
        ["pip", "download", "cowsay==6.1", "-d", str(dummy_tesseract_package)]
    )
    with open(dummy_tesseract_package / "tesseract_api.py", "w") as f:
        f.write(tesseract_api)
    with open(dummy_tesseract_package / "tesseract_requirements.txt", "w") as f:
        f.write(tesseract_requirements)

    cli_runner = CliRunner(mix_stderr=False)
    result = cli_runner.invoke(
        app,
        ["--loglevel", "debug", "build", str(dummy_tesseract_package)],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr



#!/usr/bin/env python

# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import re
import subprocess
import sys
import time
import webbrowser
from collections import defaultdict
from contextlib import nullcontext
from enum import Enum
from logging import getLogger
from pathlib import Path
from types import SimpleNamespace
from typing import Annotated, Any, NoReturn

import click
import docker
import docker.errors
import docker.models
import docker.models.containers
import docker.models.images
import typer
from jinja2 import Environment, PackageLoader, StrictUndefined
from rich.console import Console as RichConsole
from rich.table import Table as RichTable

from . import engine
from .api_parse import (
    EXPECTED_OBJECTS,
    TesseractBuildConfig,
    TesseractConfig,
    ValidationError,
    get_non_base_fields_in_tesseract_config,
)
from .exceptions import UserError
from .logs import DEFAULT_CONSOLE, set_logger

logger = getLogger("tesseract")

# Jinja2 Template Environment
ENV = Environment(
    loader=PackageLoader("tesseract_core.sdk", "templates"),
    undefined=StrictUndefined,
)


class SpellcheckedTyperGroup(typer.core.TyperGroup):
    """A Typer group that suggests similar commands if a command is not found."""

    def get_command(self, ctx: click.Context, invoked_command: str) -> Any:
        """Get a command from the Typer group, suggesting similar commands if the command is not found."""
        import difflib

        possible_commands = self.list_commands(ctx)
        if invoked_command not in possible_commands:
            close_match = difflib.get_close_matches(
                invoked_command, possible_commands, n=1, cutoff=0.6
            )
            if close_match:
                raise click.UsageError(
                    f"No such command '{invoked_command}'. Did you mean '{close_match[0]}'?",
                    ctx,
                )
        return super().get_command(ctx, invoked_command)


app = typer.Typer(
    # Make -h an alias for --help
    context_settings={"help_option_names": ["-h", "--help"]},
    name="Tesseract",
    pretty_exceptions_show_locals=False,
    cls=SpellcheckedTyperGroup,
)

# Module-wide config
state = SimpleNamespace()
state.print_user_error_tracebacks = False

# Create a list of possible commands based on the ones in api_parse (kebab-cased)
POSSIBLE_CMDS = set(
    re.sub(r"([a-z])([A-Z])", r"\1-\2", object.name).replace("_", "-").lower()
    for object in EXPECTED_OBJECTS
)
POSSIBLE_CMDS.update({"health", "openapi-schema", "check", "check-gradients"})

# All fields in TesseractConfig and TesseractBuildConfig for config override
POSSIBLE_KEYPATHS = TesseractConfig.model_fields.keys()
# Check that the only field that has nested fields is build_config
assert len(get_non_base_fields_in_tesseract_config()) == 1
POSSIBLE_BUILD_CONFIGS = TesseractBuildConfig.model_fields.keys()

# Traverse templates folder to seach for recipes
AVAILABLE_RECIPES = set()
for temp_with_path in ENV.list_templates(extensions=["py"]):
    temp_with_path = Path(temp_with_path)
    if temp_with_path.name == "tesseract_api.py" and temp_with_path.parent:
        AVAILABLE_RECIPES.add(str(temp_with_path.parent))
AVAILABLE_RECIPES = sorted(AVAILABLE_RECIPES)


class LogLevel(str, Enum):
    """Available log levels for Tesseract CLI."""

    # Must be an enum to represent a choice in Typer
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


def _validate_tesseract_name(name: str | None) -> str:
    if name is None:
        if sys.stdout.isatty() and sys.stdin.isatty():
            name = typer.prompt("Enter a name for the Tesseract")
        else:
            raise typer.BadParameter(
                "Name must be provided as an argument or interactively."
            )

    forbidden_characters: str = ":;,.@#$%^&*()[]{}<>?|\\`~"
    if any(char in forbidden_characters for char in name) or any(
        char.isspace() for char in name
    ):
        raise typer.BadParameter(
            f"Name cannot contain whitespace or any of the following characters: {forbidden_characters}"
        )
    return name


def version_callback(value: bool | None) -> None:
    """Typer callback for version option."""
    if value:
        from tesseract_core import __version__

        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def main_callback(
    loglevel: Annotated[
        LogLevel,
        typer.Option(
            help="Set the logging level. At debug level, also print tracebacks for user errors.",
            case_sensitive=False,
            show_default=True,
            metavar="LEVEL",
        ),
    ] = LogLevel.info,
    version: Annotated[
        bool | None,
        typer.Option(
            "--version",
            callback=version_callback,
            is_eager=True,
            help="Print Tesseract CLI version and exit.",
        ),
    ] = None,
) -> None:
    """Tesseract: A toolkit for re-usable, autodiff-native software components."""
    verbose_tracebacks = loglevel == LogLevel.debug
    state.print_user_error_tracebacks = verbose_tracebacks
    app.pretty_exceptions_show_locals = verbose_tracebacks

    set_logger(loglevel, catch_warnings=True, rich_format=True)


def _parse_config_override(
    options: list[str] | None,
) -> tuple[tuple[list[str], str], ...]:
    """Parse `["path1.path2.path3=value"]` into `[(["path1", "path2", "path3"], "value")]`."""
    if options is None:
        return []

    def _parse_option(option: str):
        bad_param = typer.BadParameter(
            f"Invalid config override {option} (must be `keypath=value`)",
            param_hint="config_override",
        )
        if option.count("=") != 1:
            raise bad_param

        key, value = option.split("=")
        if not key or not value:
            raise bad_param

        path = key.split(".")
        return path, value

    return tuple(_parse_option(option) for option in options)


@app.command("build")
@engine.needs_docker
def build_image(
    src_dir: Annotated[
        Path,
        typer.Argument(
            help=(
                "Source directory for the Tesseract. Must contain `tesseract_api.py` "
                "and `tesseract_config.yaml`."
            ),
            dir_okay=True,
            exists=True,
            file_okay=False,
            readable=True,
        ),
    ],
    tag: Annotated[
        str | None,
        typer.Option(
            "--tag",
            "-t",
            help="Tag for the resulting Tesseract. This can help you distinguish between "
            "Tesseracts with the same base name, like `cfd:v1` and cfd:v2`, "
            "both named `cfd` but tagged as `v1` and `v2` respectively.",
        ),
    ] = None,
    build_dir: Annotated[
        Path | None,
        typer.Option(
            help="Directory to use for the build. Defaults to a temporary directory.",
            dir_okay=True,
            file_okay=False,
            writable=True,
        ),
    ] = None,
    forward_ssh_agent: Annotated[
        bool,
        typer.Option(
            help=(
                "Forward the SSH agent to the Docker build environment. "
                "Has to be provided if requirements.txt contains private dependencies."
            ),
        ),
    ] = False,
    config_override: Annotated[
        list[str] | None,
        typer.Option(
            help=(
                "Override a configuration option in the Tesseract. "
                "Format: ``keypath=value`` where ``keypath`` is a dot-separated path to the "
                "attribute in tesseract_config.yaml. "
                "Possible keypaths are: "
                f"{', '.join(POSSIBLE_KEYPATHS)}. \n"
                "\n Possible build_config options are: "
                f"{', '.join(POSSIBLE_BUILD_CONFIGS)}. \n"
                "\nExample: ``--config-override build_config.target_platform=linux/arm64``."
            ),
            metavar="KEYPATH=VALUE",
        ),
    ] = None,
    keep_build_cache: Annotated[
        bool,
        typer.Option(
            help="Keep the Docker build cache (useful for debugging).",
            envvar="TESSERACT_KEEP_BUILD_CACHE",
        ),
    ] = False,
    generate_only: Annotated[
        bool,
        typer.Option(
            help="Only generate the build context and do not actually build the image."
        ),
    ] = False,
) -> None:
    """Build a new Tesseract from a context directory.

    The passed directory must contain the files `tesseract_api.py` and `tesseract_config.yaml`
    (can be created via `tesseract init`).

    Prints the built images as JSON array to stdout, for example: `["mytesseract:latest"]`.
    If `--generate-only` is set, the path to the build context is printed instead.
    """
    if config_override is None:
        config_override = []

    parsed_config_override = _parse_config_override(config_override)

    if generate_only:
        progress_indicator = nullcontext()
    else:
        progress_indicator = DEFAULT_CONSOLE.status(
            "[white]Processing", spinner="dots", spinner_style="white"
        )

    try:
        with progress_indicator:
            build_out = engine.build_tesseract(
                src_dir,
                tag,
                build_dir=build_dir,
                inject_ssh=forward_ssh_agent,
                config_override=parsed_config_override,
                keep_build_cache=keep_build_cache,
                generate_only=generate_only,
            )
    except docker.errors.BuildError as e:
        raise UserError(f"Error building Tesseract: {e}") from e
    except docker.errors.APIError as e:
        raise UserError(f"Docker server error: {e}") from e
    except TypeError as e:
        raise UserError(f"Input error building Tesseract: {e}") from e
    except PermissionError as e:
        raise UserError(f"Permission denied: {e}") from e
    except ValidationError as e:
        raise UserError(f"Error validating tesseract_api.py: {e}") from e

    if generate_only:
        # output is the path to the build context
        build_dir = build_out
        typer.echo(build_dir)
    else:
        # output is the built image
        image = build_out
        logger.info(f"Built image {image.short_id}, {image.tags}")
        typer.echo(json.dumps(image.tags))


@app.command("init")
def init(
    name: Annotated[
        # Guaranteed to be a string by _validate_tesseract_name
        str | None,
        typer.Option(
            help="Tesseract name as specified in tesseract_config.yaml. Will be prompted if not provided.",
            callback=_validate_tesseract_name,
            show_default=False,
        ),
    ] = None,
    target_dir: Annotated[
        Path,
        typer.Option(
            help="Path to the directory where the Tesseract API module should be created.",
            dir_okay=True,
            file_okay=False,
            writable=True,
            show_default="current directory",
        ),
    ] = Path("."),
    recipe: Annotated[
        str,
        typer.Option(
            click_type=click.Choice(AVAILABLE_RECIPES),
            help="Use a pre-configured template to initialize Tesseract API and configuration.",
        ),
    ] = "base",
) -> None:
    """Initialize a new Tesseract API module."""
    logger.info(f"Initializing Tesseract {name} in directory: {target_dir}")
    engine.init_api(target_dir, name, recipe=recipe)


def _validate_port(port: str | None) -> str:
    """Validate port input."""
    if port is None:
        return ""

    port = port.strip()
    if "-" in port:
        start, end = port.split("-")
    else:
        start = end = port

    try:
        start, end = int(start), int(end)
    except ValueError as ex:
        raise typer.BadParameter(
            (f"Port '{port}' must be single integer or a range (e.g. -p '8000-8080')."),
            param_hint="port",
        ) from ex

    if start > end:
        raise typer.BadParameter(
            (
                f"Start port '{start}' must be less than "
                f"or equal to end port '{end}' (e.g. -p '8000-8080')."
            ),
            param_hint="port",
        )

    if not (1025 <= start <= 65535) or not (1025 <= end <= 65535):
        raise typer.BadParameter(
            f"Ports '{port}' must be between 1025 and 65535.",
            param_hint="port",
        )
    return port


@app.command("serve")
@engine.needs_docker
def serve(
    image_names: Annotated[
        list[str],
        typer.Argument(..., help="One or more Tesseract image names"),
    ],
    volume: Annotated[
        list[str] | None,
        typer.Option(
            "-v",
            "--volume",
            help="Bind mount a volume in all Tesseracts, in Docker format: source:target[:ro|rw]",
            metavar="source:target",
            show_default=False,
        ),
    ] = None,
    port: Annotated[
        str | None,
        typer.Option(
            "--port",
            "-p",
            help="Optional port/port range to serve the Tesseract on (e.g. -p '8080-8082'). "
            "Port must be between 1025 and 65535.",
            callback=_validate_port,
        ),
    ] = None,
    gpus: Annotated[
        list[str] | None,
        typer.Option(
            "--gpus",
            metavar="'all' | int",
            help=(
                "IDs of host Nvidia GPUs to make available in the Tesseract. "
                "You can use all GPUs via `--gpus all` or pass (multiple) IDs: `--gpus 0 --gpus 1`."
            ),
        ),
    ] = None,
) -> None:
    """Serve one or more Tesseract images.

    A successful serve command will display on standard output a JSON object
    with the Docker Compose project ID, which is required to run the teardown
    command, as well as a list of all containers spawned and their respective
    ports.
    """
    if port and len(image_names) > 1:
        # TODO: Docker compose with multiple ports is not supported until
        # docker/compose#7188 is resolved.
        raise typer.BadParameter(
            (
                "Port specification only works if 1 Tesseract is being served."
                f"Currently serving `{len(image_names)}` Tesseracts."
            ),
            param_hint="image_names",
        )

    try:
        project_id = engine.serve(image_names, port, volume, gpus)
        containers = engine.project_containers(project_id)
        _display_container_meta(containers)
        logger.info(
            f"Docker Compose Project ID, use it with 'tesseract teardown' command: {project_id}"
        )

        project_meta = {"project_id": project_id, "containers": []}
        for container in containers:
            project_meta["containers"].append(
                {"name": container.name, "port": _get_container_host_port(container)}
            )

        json_info = json.dumps(project_meta)
        typer.echo(json_info, nl=False)

    except ValueError as ex:
        raise typer.BadParameter(f"{ex}", param_hint="image_names") from ex
    except RuntimeError as ex:
        raise UserError(
            f"Internal Docker error occurred while serving Tesseracts: {ex}"
        ) from ex


@app.command("list")
@engine.needs_docker
def list_tesseract_images() -> None:
    """Display all Tesseract images."""
    tesseract_images = engine.get_tesseract_images()
    _display_tesseract_image_meta(tesseract_images)


@app.command("ps")
@engine.needs_docker
def list_tesseract_containers() -> None:
    """Display all Tesseract containers."""
    tesseract_containers = engine.get_tesseract_containers()
    _display_tesseract_containers_meta(tesseract_containers)


def _display_tesseract_image_meta(
    docker_assets: list[docker.models.images.Image],
) -> None:
    """Display Tesseract image metadata."""
    table = RichTable("ID", "Tags", "Name", "Version", "Description")
    for asset in docker_assets:
        tesseract_vals = _get_tesseract_env_vals(asset)
        if tesseract_vals:
            table.add_row(
                # Checksum Type + First 12 Chars of ID
                asset.id[:19],
                str(asset.attrs["RepoTags"]),
                tesseract_vals["TESSERACT_NAME"],
                tesseract_vals["TESSERACT_VERSION"],
                tesseract_vals.get("TESSERACT_DESCRIPTION", "").replace("\n", " "),
            )
    RichConsole().print(table)


def _display_tesseract_containers_meta(
    docker_assets: list[docker.models.containers.Container],
) -> None:
    """Display Tesseract containers metadata."""
    table = RichTable("ID", "Name", "Version", "Host Port", "Project ID", "Description")
    docker_compose_projects = _docker_compose_projects()

    for asset in docker_assets:
        tesseract_vals = _get_tesseract_env_vals(asset)
        if tesseract_vals:
            tesseract_project = _find_tesseract_project(asset, docker_compose_projects)
            table.add_row(
                asset.id[:12],
                tesseract_vals["TESSERACT_NAME"],
                tesseract_vals["TESSERACT_VERSION"],
                _get_container_host_port(asset),
                tesseract_project,
                tesseract_vals.get("TESSERACT_DESCRIPTION", "").replace("\\n", " "),
            )
    RichConsole().print(table)


def _get_tesseract_env_vals(
    docker_asset: docker.models.images.Image | docker.models.containers.Container,
) -> dict:
    """Convert Tesseract environment variables from list to dictionary."""
    env_vals = [s for s in docker_asset.attrs["Config"]["Env"] if "TESSERACT_" in s]
    return {item.split("=")[0]: item.split("=")[1] for item in env_vals}


def _find_tesseract_project(
    tesseract: docker.models.containers.Container,
    docker_compose_projects: defaultdict[str, list],
) -> str:
    """Find the Tesseract Project ID for a given tesseract."""
    tesseract_id = tesseract.id[:12]

    for project, containers in docker_compose_projects.items():
        if tesseract_id in containers:
            return project

    return "Unknown"


def _docker_compose_projects() -> defaultdict[str, list]:
    """List Docker Compose projects.

    Build a dictionary {project_name: [container ID]}.
    """
    proc = subprocess.run(
        ["docker", "compose", "ls", "--format", "json"],
        check=True,
        capture_output=True,
    )

    out = proc.stdout.decode().strip()
    compose_projects = json.loads(out)

    projects_map = defaultdict(list)

    for project in compose_projects:
        project_name = project["Name"]

        proc = subprocess.run(
            [
                "docker",
                "compose",
                "--project-name",
                project_name,
                "ps",
                "--format",
                "json",
            ],
            check=True,
            capture_output=True,
        )
        # This command outputs a series of JSON documents, one for each
        # container, instead of a JSON with a list of entries.
        compose_ps = proc.stdout.decode().strip().split("\n")
        for asset in compose_ps:
            metadata = json.loads(asset)
            projects_map[project_name].append(metadata["ID"])

    return projects_map


@app.command("apidoc")
@engine.needs_docker
def apidoc(
    image_name: Annotated[
        str,
        typer.Argument(..., help="Tesseract image name"),
    ],
    browser: Annotated[
        bool,
        typer.Option(help="Open the browser after serving the OpenAPI schema"),
    ] = True,
) -> None:
    """Serve the OpenAPI schema for a Tesseract."""
    project_id = None

    try:
        project_id = engine.serve([image_name])
        container = engine.project_containers(project_id)[0]
        host_port = _get_container_host_port(container)
        url = f"http://localhost:{host_port}/docs"
        logger.info(f"Serving OpenAPI docs for Tesseract {image_name} at {url}")
        logger.info("  Press Ctrl+C to stop")
        if browser:
            webbrowser.open(url)
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            return
    finally:
        if project_id is not None:
            engine.teardown(project_id)


def _display_container_meta(
    containers: list[docker.models.containers.Container],
) -> None:
    """Display container metadata."""
    for container in containers:
        logger.info(f"Container ID: {container.id}")
        logger.info(f"Name: {container.name}")
        entrypoint = container.attrs["Config"]["Entrypoint"]
        logger.info(f"Entrypoint: {entrypoint}")
        port_key = next(iter(container.ports))
        host_port = container.ports[port_key][0]["HostPort"]
        logger.info(f"View Tesseract: http://localhost:{host_port}/docs")


def _get_container_host_port(container: docker.models.containers.Container):
    port_key = next(iter(container.ports))
    host_port = container.ports[port_key][0]["HostPort"]
    return host_port


@app.command("teardown")
@engine.needs_docker
def teardown(
    project_ids: Annotated[
        list[str] | None,
        typer.Argument(..., help="Docker Compose project IDs"),
    ] = None,
    tear_all: Annotated[
        bool,
        typer.Option(
            "--all",
            help="Tear down all Tesseracts currently being served.",
        ),
    ] = False,
) -> None:
    """Tear down one or more Tesseracts that were previously started with `tesseract serve`.

    One or more Tesseract project ids must be specified unless `--all` is set.
    """
    if not project_ids and not tear_all:
        raise typer.BadParameter(
            "Either project IDs or --all flag must be provided",
            param_hint="project_ids",
        )

    if project_ids and tear_all:
        raise typer.BadParameter(
            "Either project IDs or --all flag must be provided, but not both",
            param_hint="project_ids",
        )
    if tear_all:
        project_ids = []
        docker_compose_projects = _docker_compose_projects()
        for project, _ in docker_compose_projects.items():
            if "tesseract-" in project:
                project_ids.append(project)

    for project_id in project_ids:
        try:
            engine.teardown(project_id)
        except ValueError as ex:
            raise UserError(
                f"Input error occurred while tearing down Tesseracts: {ex}"
            ) from ex
        except RuntimeError as ex:
            raise UserError(
                f"Internal Docker error occurred while tearing down Tesseracts: {ex}"
            ) from ex
        logger.info(
            f"Tesseracts are shutdown for Docker Compose project ID: {project_id}"
        )


def _sanitize_error_output(error_output: str, tesseract_image: str) -> str:
    """Remove references to tesseract-runtime and unavailable commands from error output."""
    # Replace references to tesseract-runtime with tesseract run
    error_output = re.sub(
        r"Try 'tesseract-runtime",
        f"Try 'tesseract run {tesseract_image}",
        error_output,
    )

    error_output = re.sub(
        r"Usage: tesseract-runtime",
        f"Usage: tesseract run {tesseract_image}",
        error_output,
    )

    # Hide commands in help strings that users are not supposed to use via tesseract run
    error_output = re.sub(
        r"^│\s+(serve|health)\s+.*?$\n",
        "",
        error_output,
        flags=re.MULTILINE,
    )

    return error_output


@app.command(
    "run",
    # We need to ignore unknown options to forward the args to the Tesseract container
    context_settings={"ignore_unknown_options": True},
    # We implement --help manually to forward the help of the Tesseract container
    add_help_option=False,
)
@engine.needs_docker
def run_container(
    context: click.Context,
    tesseract_image: Annotated[
        str,
        typer.Argument(help="Tesseract image name"),
    ],
    cmd: Annotated[
        str,
        typer.Argument(help="Tesseract command to run"),
    ] = "",
    args: Annotated[
        list[str] | None,
        typer.Argument(help="Arguments for the command"),
    ] = None,
    volume: Annotated[
        list[str] | None,
        typer.Option(
            "-v",
            "--volume",
            help="Bind mount a volume, in Docker format: source:target.",
            metavar="source:target",
            show_default=False,
        ),
    ] = None,
    gpus: Annotated[
        list[str] | None,
        typer.Option(
            "--gpus",
            metavar="'all' | int",
            help=(
                "IDs of host GPUs to make available in the tesseract. "
                "You can use all GPUs via `--gpus all` or pass (multiple) IDs: `--gpus 0 --gpus 1`."
            ),
        ),
    ] = None,
) -> None:
    """Execute a command in a Tesseract.

    This command starts a Tesseract instance and executes the given
    command.
    """
    if args is None:
        args = []

    if cmd == "serve":
        logger.error(
            "You should not serve tesseracts via "
            "`tesseract run <tesseract-name> serve`. "
            "Use `tesseract serve <tesseract-name>` instead."
        )
        raise typer.Exit(1)

    help_args = {"-h", "--help"}

    # When called as `tesseract run --help` -> show generic help
    if tesseract_image in help_args:
        context.get_help()
        return

    invoke_help = any(arg in help_args for arg in args) or cmd in help_args

    if (not cmd or cmd not in POSSIBLE_CMDS) and not invoke_help:
        if not cmd:
            error_string = f"Command is required. Are you sure your Tesseract image name is `{tesseract_image}`?\n"
        else:
            error_string = f"Command `{cmd}` does not exist. \n"

        error_string += (
            f"\nRun `tesseract run {tesseract_image} --help` for more information.\n"
        )

        error_string = (
            error_string + f"\nPossible commands are: {', '.join(POSSIBLE_CMDS)}"
        )
        raise typer.BadParameter(error_string, param_hint="cmd")

    try:
        result_out, result_err = engine.run_tesseract(
            tesseract_image, cmd, args, volumes=volume, gpus=gpus
        )

    except docker.errors.ImageNotFound as e:
        raise UserError(
            "Tesseract image not found. "
            f"Are you sure your tesseract image name is {tesseract_image}?\n\n{e}"
        ) from e

    except (
        docker.errors.APIError,
        docker.errors.ContainerError,
    ) as e:
        if "No such command" in str(e):
            error_string = f"Error running Tesseract '{tesseract_image}' \n\n Error: Unimplemented command '{cmd}'.  "
        else:
            error_string = _sanitize_error_output(
                f"Error running Tesseract. \n\n{e}", tesseract_image
            )

        raise UserError(error_string) from e

    if invoke_help:
        result_err = _sanitize_error_output(result_err, tesseract_image)

    typer.echo(result_err, err=True, nl=False)
    typer.echo(result_out, nl=False)


def entrypoint() -> NoReturn:
    """Entrypoint for the Tesseract CLI."""
    try:
        result = app()
    except UserError as e:
        logger.error(f"{e}", exc_info=state.print_user_error_tracebacks)
        result = 1
    except Exception as e:
        logger.critical(f"Uncaught error: {e}", exc_info=True)
        result = 2

    if result > 0:
        logger.critical("Aborting")

    raise SystemExit(result)


# Expose the underlying click object for doc generation
typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    entrypoint()



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import traceback

from typer.testing import CliRunner

from tesseract_core.sdk.cli import app


def image_exists(client, image_name):
    # Docker images may be prefixed with the registry URL
    return any(
        tag.split("/")[-1] == image_name
        for img in client.images.list()
        for tag in img.tags
    )


def print_debug_info(result):
    """Print debug info from result of a CLI command if it failed."""
    if result.exit_code == 0:
        return
    print(result.output)
    if result.exc_info:
        traceback.print_exception(*result.exc_info)


def build_tesseract(sourcedir, image_name, tag=None, build_retries=3):
    cli_runner = CliRunner(mix_stderr=False)

    build_args = [
        "--loglevel",
        "debug",
        "build",
        str(sourcedir),
        "--config-override",
        f"name={image_name}",
    ]

    if tag is not None:
        build_args.extend(["--tag", tag])
        image_name = f"{image_name}:{tag}"
    else:
        image_name = f"{image_name}:latest"

    for _ in range(build_retries):
        result = cli_runner.invoke(
            app,
            build_args,
            catch_exceptions=False,
        )
        # Retry if the build fails with EOF error (connectivity issue)
        # See https://github.com/docker/buildx/issues/2064
        is_expected_err = "error reading from server: EOF" in result.output
        if not is_expected_err:
            break

    print_debug_info(result)
    assert result.exit_code == 0, result.exception

    image_tags = json.loads(result.stdout.strip())
    assert image_name in image_tags
    return image_tags[0]



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import base64
import json
import os
import subprocess
import sys
import time
from textwrap import dedent

import numpy as np
import pytest
import requests
from fastapi.testclient import TestClient

from tesseract_core.runtime.serve import create_rest_api

test_input = {
    "a": [1.0, 2.0, 3.0],
    "b": [1, 1, 1],
    "s": 2.5,
}


def array_from_json(json_data):
    encoding = json_data["data"]["encoding"]
    if encoding == "base64":
        decoded_buffer = base64.b64decode(json_data["data"]["buffer"])
        array = np.frombuffer(decoded_buffer, dtype=json_data["dtype"]).reshape(
            json_data["shape"]
        )
    elif encoding == "json":
        array = np.array(json_data["data"]["buffer"], dtype=json_data["dtype"]).reshape(
            json_data["shape"]
        )

    return array


def model_to_json(model):
    return json.loads(model.model_dump_json())


@pytest.fixture
def mock_client(dummy_tesseract_module):
    """Mock an http client."""
    rest_api = create_rest_api(dummy_tesseract_module)
    return TestClient(rest_api)


@pytest.mark.parametrize(
    "format",
    [
        "json",
        "json+base64",
        pytest.param("json+binref", marks=pytest.mark.xfail),  # FIXME
        "msgpack",
    ],
)
def test_create_rest_api_apply_endpoint(mock_client, dummy_tesseract_module, format):
    """Test we can get an Apply endpoint from generated API."""
    test_inputs = dummy_tesseract_module.InputSchema.model_validate(test_input)

    response = mock_client.post(
        "/apply",
        json={"inputs": model_to_json(test_inputs)},
        headers={"Accept": f"application/{format}"},
    )

    assert response.status_code == 200, response.text

    if format in {"json", "json+base64"}:
        result = array_from_json(response.json()["result"])
        assert np.array_equal(result, np.array([3.5, 6.0, 8.5]))
    elif format == "msgpack":
        assert (
            response.content
            == b"\x81\xa6result\x85\xc4\x02nd\xc3\xc4\x04type\xa3<f4\xc4\x04kind\xc4\x00\xc4\x05shape\x91\x03\xc4\x04data\xc4\x0c\x00\x00`@\x00\x00\xc0@\x00\x00\x08A"  # noqa: E501
        )
    elif format == "json+binref":
        raise NotImplementedError()


def test_create_rest_api_jacobian_endpoint(mock_client, dummy_tesseract_module):
    """Test we can get a Jacobian endpoint from generated API."""
    test_inputs = dummy_tesseract_module.InputSchema.model_validate(test_input)

    response = mock_client.post(
        "/jacobian",
        json={
            "inputs": model_to_json(test_inputs),
            "jac_inputs": ["a", "b"],
            "jac_outputs": ["result"],
        },
    )

    assert response.status_code == 200, response.text
    result = response.json()
    expected = dummy_tesseract_module.jacobian(test_inputs, ["a", "b"], ["result"])

    assert result.keys() == expected.keys()
    assert np.array_equal(
        array_from_json(result["result"]["a"]), expected["result"]["a"]
    )


def test_create_rest_api_generates_health_endpoint(mock_client):
    """Test we can get health endpoint from generated API."""
    response = mock_client.get("/health")
    assert response.json() == {"status": "ok"}


def test_get_input_schema(mock_client):
    response = mock_client.get("/input_schema")

    assert response.status_code == 200, response.text


def test_get_output_schema(mock_client):
    response = mock_client.get("/output_schema")

    assert response.status_code == 200, response.text


def test_post_abstract_eval(mock_client):
    payload = {
        "inputs": {
            "a": {"dtype": "float64", "shape": [4]},
            "b": {"dtype": "float64", "shape": [4]},
            "s": 1.0,
            "normalize": False,
        }
    }
    response = mock_client.post("/abstract_eval", json=payload)

    assert response.status_code == 200, response.text
    assert response.json() == {"result": {"shape": [4], "dtype": "float64"}}


def test_post_abstract_eval_throws_validation_errors(mock_client):
    response = mock_client.post("/abstract_eval", json={"what": {"is": "this"}})

    assert response.status_code == 422, response.text
    errors = response.json()["detail"]
    error_types = [e["type"] for e in errors]

    assert "missing" in error_types
    assert "extra_forbidden" in error_types


def test_get_openapi_schema(mock_client):
    response = mock_client.get("/openapi.json")

    assert response.status_code == 200, response.text
    assert response.json()["info"]["title"] == "Tesseract"
    assert response.json()["paths"]


def test_threading_sanity(tmpdir, free_port):
    """Test with a Tesseract that requires to be run in the main thread.

    This is important so we don't require users to be aware of threading issues.
    """
    TESSERACT_API = dedent(
        """
    import threading
    from pydantic import BaseModel

    assert threading.current_thread() == threading.main_thread()

    class InputSchema(BaseModel):
        pass

    class OutputSchema(BaseModel):
        pass

    def apply(input: InputSchema) -> OutputSchema:
        assert threading.current_thread() == threading.main_thread()
        return OutputSchema()

    def abstract_eval(abstract_inputs: dict) -> dict:
        pass
    """
    )

    api_file = tmpdir / "tesseract_api.py"

    with open(api_file, "w") as f:
        f.write(TESSERACT_API)

    # We can't run the server in the same process because it will use threading under the hood
    # so we need to spawn a new process instead
    try:
        proc = subprocess.Popen(
            [
                sys.executable,
                "-c",
                "from tesseract_core.runtime.serve import serve; "
                f"serve(host='localhost', port={free_port}, num_workers=1)",
            ],
            env=dict(os.environ, TESSERACT_API_PATH=api_file),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # wait for server to start
        timeout = 10.0
        while True:
            try:
                response = requests.get(f"http://localhost:{free_port}/health")
            except requests.exceptions.ConnectionError:
                pass
            else:
                if response.status_code == 200:
                    break

            time.sleep(0.1)
            timeout -= 0.1

            if timeout < 0:
                raise TimeoutError("Server did not start in time")

        response = requests.post(
            f"http://localhost:{free_port}/apply", json={"inputs": {}}
        )
        assert response.status_code == 200, response.text

    finally:
        proc.terminate()
        stdout, stderr = proc.communicate()
        print(stdout.decode())
        print(stderr.decode())
        proc.wait(timeout=5)
